/*******************************************************************************
NAME                     ALBERS CONICAL EQUAL-AREA 

PURPOSE:    Transforms input Easting and Northing to longitude and
        latitude for the Albers Conical Equal Area projection.  The
        Easting and Northing values are in meters.  The longitude
        and latitude values are in radians.

ALGORITHM REFERENCES

1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.

2.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.
*******************************************************************************/
#include <stdlib.h>
#include "cproj.h"
#include "gctp.h"
#include "local.h"

/* Variables common to all subroutines in this code file
  -----------------------------------------------------*/
/* Structure to hold the setup data relevant to this projection */
struct albers_proj
{
    double r_major;        /* major axis */
    double r_minor;        /* minor axis */
    double c;              /* constant c */
    double e3;             /* eccentricity */
    double es;             /* eccentricity squared */
    double rh;             /* height above ellipsoid */
    double ns0;            /* ratio between meridians */
    double center_lon;     /* center longitude */
    double false_easting;  /* x offset in meters */
    double false_northing; /* y offset in meters */
    double latitude1;      /* Latitude 1, parallel */
    double latitude2;      /* Latitude 2, parallel */
    double center_lat;     /* center lat */
};

/*****************************************************************************
Name: print_info

Purpose: Prints a summary of information about this projection.

Returns:
    nothing

*****************************************************************************/
static void print_info
(
    const TRANSFORMATION *trans
)
{
    struct albers_proj *cache_ptr = (struct albers_proj *)trans->cache;

    /* Report parameters to the user
      -----------------------------*/
    gctp_print_title("ALBERS CONICAL EQUAL-AREA");
    gctp_print_radius2(cache_ptr->r_major, cache_ptr->r_minor);
    gctp_print_stanparl(cache_ptr->latitude1, cache_ptr->latitude2);
    gctp_print_cenlonmer(cache_ptr->center_lon);
    gctp_print_origin(cache_ptr->center_lat);
    gctp_print_offsetp(cache_ptr->false_easting, cache_ptr->false_northing);
}

/*******************************************************************************
Name: calc_phi1

Purpose: Function to compute phi1, the latitude for the inverse of the
   Albers Conical Equal-Area projection.

Returns:
    GCTP_SUCCESS or GCTP_ERROR

*******************************************************************************/
static int calc_phi1
(
    double eccent,  /* I: Eccentricity angle in radians */
    double qs,      /* I: Angle in radians */
    double *phi1    /* O: Calculated phi1 value */
)
{
    double eccnts;
    double dphi;
    double con;
    double com;
    double sinpi;
    double cospi;
    double phi;
    long i;

    phi = asinz(.5 * qs);
    if (eccent < EPSLN) 
    {
        *phi1 = phi;
        return GCTP_SUCCESS;
    }

    eccnts = eccent * eccent; 
    for (i = 1; i <= 25; i++)
    {
        sincos(phi,&sinpi,&cospi);
        con = eccent * sinpi; 
        com = 1.0 - con * con;
        dphi = .5 * com * com / cospi * (qs / (1.0 - eccnts) - sinpi / com + 
                .5 / eccent * log ((1.0 - con) / (1.0 + con)));
        phi = phi + dphi;
        if (fabs(dphi) <= 1e-7)
        {
            *phi1 = phi;
            return GCTP_SUCCESS;
        }
    }
    GCTP_PRINT_ERROR("Convergence error");
    *phi1 = 0.0;
    return GCTP_ERROR;
}

/*******************************************************************************
Name: calc_small_q

Purpose: Function to compute constant small q which is the radius of a 
   parallel of latitude, phi, divided by the semimajor axis. 

Returns:
    GCTP_SUCCESS or GCTP_ERROR

*******************************************************************************/
static double calc_small_q
(
    double eccent,
    double sinphi
)
{
    double con;

    if (eccent > 1.0e-7)
    {
        con = eccent * sinphi;
        return (( 1.0- eccent * eccent) * (sinphi /(1.0 - con * con) - 
                (.5/eccent) * log((1.0 - con)/(1.0 + con))));
    }
    else
        return 2.0 * sinphi;
}

/*******************************************************************************
Name: common_init

Purpose: Initialization routine for initializing the projection information
    that is common to both the forward and inverse transformations.

Returns:
    GCTP_SUCCESS or GCTP_ERROR

*******************************************************************************/
static int common_init
(
    TRANSFORMATION *trans /* I/O: transformation to initialize */
)
{
    double r_major;             /* major axis */
    double r_minor;             /* minor axis */
    double radius;
    double latitude1;           /* first standard parallel */
    double latitude2;           /* second standard parallel */
    double center_lon;          /* center longitude */
    double center_lat;          /* center latitude */
    double false_easting;       /* x offset in meters */
    double false_northing;      /* y offset in meters */

    double sin_po,cos_po;       /* sine and cos values */
    double c;
    double con;                 /* temporary variable */
    double es;                  /* eccentricity squared */
    double e3;                  /* eccentricity */
    double ms1;                 /* small m 1 */
    double ms2;                 /* small m 2 */
    double ns0;                 /* ratio between meridians */
    double qs0;                 /* small q 0 */
    double qs1;                 /* small q 1 */
    double qs2;                 /* small q 2 */
    double rh;                  /* height above ellipsoid */
    double temp;                /* temporary variable */

    const GCTP_PROJECTION *proj = &trans->proj;
    struct albers_proj *cache = NULL;
    int spheroid = proj->spheroid;

    gctp_get_spheroid(spheroid, proj->parameters, &r_major, &r_minor, &radius);
    false_easting  = proj->parameters[6];
    false_northing = proj->parameters[7];

    /* Lat parallel 1 to use */
    if (gctp_dms2degrees(proj->parameters[2], &latitude1) != GCTP_SUCCESS)
    {
        GCTP_PRINT_ERROR("Error converting first standard parallel in "
            "parameter 2 from DMS to degrees: %f", proj->parameters[2]);
        return GCTP_ERROR;
    }
    latitude1 *= 3600 * S2R;

    /* Lat parallel 2 to use */
    if (gctp_dms2degrees(proj->parameters[3], &latitude2) != GCTP_SUCCESS)
    {
        GCTP_PRINT_ERROR("Error converting second standard parallel in "
            "parameter 3 from DMS to degrees: %f", proj->parameters[3]);
        return GCTP_ERROR;
    }
    latitude2 *= 3600 * S2R;

    /* Lat origin */
    if (gctp_dms2degrees(proj->parameters[5], &center_lat) != GCTP_SUCCESS)
    {
        GCTP_PRINT_ERROR("Error converting center latitude in parameter 4 "
            "from DMS to degrees: %f", proj->parameters[5]);
        return GCTP_ERROR;
    }
    center_lat *= 3600 * S2R;

    /* Center longitude (central meridian) */
    if (gctp_dms2degrees(proj->parameters[4], &center_lon) != GCTP_SUCCESS)
    {
        GCTP_PRINT_ERROR("Error converting center longitude in parameter 4 "
            "from DMS to degrees: %f", proj->parameters[4]);
        return GCTP_ERROR;
    }
    center_lon *= 3600 * S2R;

    /* Allocate a structure for the cached info */
    cache = malloc(sizeof(*cache));
    if (!cache)
    {
        GCTP_PRINT_ERROR("Error allocating memory for cache buffer");
        return GCTP_ERROR;
    }
    trans->cache = cache;

    /* Save the information to the cache */
    cache->latitude1 = latitude1;
    cache->latitude2 = latitude2;
    cache->r_major = r_major;
    cache->r_minor = r_minor;
    cache->false_easting = false_easting;
    cache->false_northing = false_northing;
    temp = r_minor / r_major;
    es = 1.0 - SQUARE(temp);
    cache->es = es;
    e3 = sqrt(es);
    cache->e3 = sqrt(es);
    cache->center_lat = center_lat;
    cache->center_lon = center_lon;

    sincos(latitude1, &sin_po, &cos_po);
    con = sin_po;

    ms1 = gctp_calc_small_radius(e3, sin_po, cos_po);
    qs1 = calc_small_q(e3,sin_po);

    sincos(latitude2, &sin_po, &cos_po);

    ms2 = gctp_calc_small_radius(e3, sin_po, cos_po);
    qs2 = calc_small_q(e3, sin_po);

    sincos(center_lat, &sin_po, &cos_po);

    qs0 = calc_small_q(e3, sin_po);

    if (fabs(latitude1 - latitude2) > EPSLN)
       ns0 = (ms1 * ms1 - ms2 *ms2)/ (qs2 - qs1);
    else
       ns0 = con;
    c = ms1 * ms1 + ns0 * qs1;
    rh = r_major * sqrt(c - ns0 * qs0)/ns0;

    /* constant */
    cache->c = c;
    /* Height above ellipsoid */
    cache->rh = rh;
    /* Ration between meridians */
    cache->ns0 = ns0;

    trans->print_info = print_info;

    return GCTP_SUCCESS;
}

/*****************************************************************************
Name: inverse_transform

Purpose: Transforms X,Y to lat,long

Returns:
    GCTP_SUCCESS or GCTP_ERROR

*****************************************************************************/
static int inverse_transform 
(
    const TRANSFORMATION *trans, /* I: transformation information */
    double x,       /* I: X projection coordinate */
    double y,       /* I: Y projection coordinate */
    double *lon,    /* O: Longitude */
    double *lat     /* O: Latitude */
)
{
    struct albers_proj *cache_ptr = (struct albers_proj *)trans->cache;
    double rh1;         /* height above ellipsoid */
    double qs;          /* function q */
    double con;         /* temporary sign value */
    double theta;       /* angle */


    x -= cache_ptr->false_easting;
    y = cache_ptr->rh - y + cache_ptr->false_northing;
    if (cache_ptr->ns0 >= 0)
    {
        rh1 = sqrt(x * x + y * y);
        con = 1.0;
    }
    else
    {
       rh1 = -sqrt(x * x + y * y);
       con = -1.0;
    }
    theta = 0.0;
    if (rh1 != 0.0)
        theta = atan2(con * x, con * y);
    con = rh1 * cache_ptr->ns0 / cache_ptr->r_major;
    qs = (cache_ptr->c - con * con) / cache_ptr->ns0;
    if (cache_ptr->e3 >= 1e-10)
    {
        con = 1 - .5 * (1.0 - cache_ptr->es) * log((1.0 - cache_ptr->e3) / 
              (1.0 + cache_ptr->e3))/cache_ptr->e3;
        if (fabs(fabs(con) - fabs(qs)) > .0000000001 )
        {
            if (calc_phi1(cache_ptr->e3, qs, lat) != GCTP_SUCCESS)
                return GCTP_ERROR;
        }
        else
        {
            if (qs >= 0)
                *lat = .5 * PI;
            else
                *lat = -.5 * PI;
        }
    }
    else
    {
        if (calc_phi1(cache_ptr->e3, qs, lat) != GCTP_SUCCESS)
            return GCTP_ERROR;
    }

    *lon = adjust_lon(theta/cache_ptr->ns0 + cache_ptr->center_lon);

    return GCTP_SUCCESS;
}

/*****************************************************************************
Name: forward_transform

Purpose: Transforms lat,long to albers X,Y

Returns:
    GCTP_SUCCESS or GCTP_ERROR

*****************************************************************************/
static int forward_transform
(
    const TRANSFORMATION *trans, /* I: transformation information */
    double lon,         /* I: Longitude */
    double lat,         /* I: Latitude */
    double *x,          /* O: X projection coordinate */
    double *y           /* O: Y projection coordinate */
)

{
    struct albers_proj *cache_ptr = (struct albers_proj *)trans->cache;

    double sin_phi;     /* sine value */
    double cos_phi;     /* cos value */
    double qs;          /* small q */
    double theta;       /* angle */
    double rh1;         /* height above ellipsoid */

    sincos(lat, &sin_phi, &cos_phi);
    qs = calc_small_q(cache_ptr->e3, sin_phi);
    rh1 = cache_ptr->r_major * sqrt(cache_ptr->c - cache_ptr->ns0 * qs) /
          cache_ptr->ns0;
    theta = cache_ptr->ns0 * adjust_lon(lon - cache_ptr->center_lon);
    *x = rh1 * sin(theta) + cache_ptr->false_easting;
    *y = cache_ptr->rh - rh1 * cos(theta) + cache_ptr->false_northing;

    return GCTP_SUCCESS;
}

/*****************************************************************************
Name: gctp_albers_inverse_init

Purpose: Initializes the inverse albers transformation

Returns:
    GCTP_SUCCESS or GCTP_ERROR

*****************************************************************************/
int gctp_albers_inverse_init
(
    TRANSFORMATION *trans
)
{
    /* Call the common routine used for the forward and inverse init */
    if (common_init(trans) != GCTP_SUCCESS)
    {
        GCTP_PRINT_ERROR(
            "Error initializing albers inverse projection");
        return GCTP_ERROR;
    }

    trans->transform = inverse_transform;

    return GCTP_SUCCESS;
}

/*****************************************************************************
Name: gctp_albers_forward_init

Purpose: Initializes the forward albers transformation

Returns:
    GCTP_SUCCESS or GCTP_ERROR

*****************************************************************************/
int gctp_albers_forward_init
(
    TRANSFORMATION *trans
)
{
    /* Call the common routine used for the forward and inverse init */
    if (common_init(trans) != GCTP_SUCCESS)
    {
        GCTP_PRINT_ERROR(
            "Error initializing albers forward projection");
        return GCTP_ERROR;
    }

    trans->transform = forward_transform;

    return GCTP_SUCCESS;
}
