/**************************************************************************
 * Parks-McClellan algorithm for FIR filter design (C version)
 *-------------------------------------------------
 *  Copyright (C) 1995  Jake Janovetz (janovetz@coewl.cen.uiuc.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *************************************************************************/

/***************************************************************************
 * Test program for the remez() function.  Sends appropriate arguments to
 * remez() to generate a filter.  Then, compares the resulting coefficients
 * to expected baseline results.
 **************************************************************************/

#include "remez.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#define NUM_TAPS 104
#define NUM_BANDS 5

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) > (b) ? (b) : (a))

static int significant_figures = 8;

static double accuracy(int significant_figures, double largest);

int doubles_are_equal(const double tried, const double expected) 
{
    return max(tried, expected) - min(tried, expected) < 
        accuracy(significant_figures, max(fabs(tried), fabs(expected)));
}

static double accuracy(int significant_figures, double largest) 
{
    return pow(10, 1 + (int)log10(largest) - significant_figures);
}

int main()
{
   double *weights, *desired, *bands;
   double *h;
   int i;
   int status; /* Function return status */
   double expected_results[104];
   int results_matched = 1; /* Flag to indicate if actual results matched
       expected results */

   /* Initialize the expected results */
   expected_results[0] =  0.00027151056242530401;
   expected_results[1] =  0.00222989309130521094;
   expected_results[2] =  0.00091345773396801701;
   expected_results[3] = -0.00312243468817884279;
   expected_results[4] = -0.00016728479105140539;
   expected_results[5] = -0.00002202658825579365;
   expected_results[6] = -0.00456747888272199080;
   expected_results[7] = -0.00063717027408990740;
   expected_results[8] =  0.00305927679687127392;
   expected_results[9] =  0.00119935104891564755;
   expected_results[10] = -0.00484279055650574299;
   expected_results[11] = -0.00040579843413544522;
   expected_results[12] =  0.01389581931260174506;
   expected_results[13] =  0.00096535883615447675;
   expected_results[14] = -0.00441659222550526767;
   expected_results[15] =  0.01255794680690700688;
   expected_results[16] = -0.00176665961306401976;
   expected_results[17] = -0.01695735115373797272;
   expected_results[18] = -0.00878876776373385508;
   expected_results[19] =  0.00849548530949226915;
   expected_results[20] =  0.00290880442387733175;
   expected_results[21] = -0.02907374717387719784;
   expected_results[22] =  0.00305755210718031091;
   expected_results[23] =  0.02327368288411046579;
   expected_results[24] = -0.01903462878624252730;
   expected_results[25] =  0.00775974745641534649;
   expected_results[26] =  0.04096041084703053864;
   expected_results[27] =  0.01854924814461966776;
   expected_results[28] = -0.02165135036998024759;
   expected_results[29] = -0.02261690245306222438;
   expected_results[30] =  0.03869181896686624267;
   expected_results[31] = -0.01250032551907871887;
   expected_results[32] = -0.05538408271236267694;
   expected_results[33] =  0.02643447567385441566;
   expected_results[34] = -0.00795068167375006553;
   expected_results[35] = -0.06001906385307194391;
   expected_results[36] = -0.02428828698861155641;
   expected_results[37] =  0.04786291183903680635;
   expected_results[38] =  0.07474403970895460159;
   expected_results[39] = -0.04032311897593984090;
   expected_results[40] =  0.01301250123477310644;
   expected_results[41] =  0.08689468846600580110;
   expected_results[42] = -0.04191579352767203781;
   expected_results[43] =  0.00348155518956557580;
   expected_results[44] =  0.05901165657804177678;
   expected_results[45] =  0.03928220651181194012;
   expected_results[46] = -0.08413108928807125353;
   expected_results[47] = -0.27751259202078676358;
   expected_results[48] =  0.01305019940644617724;
   expected_results[49] =  0.04202636527909262010;
   expected_results[50] = -0.18748886006595252840;
   expected_results[51] =  0.34508248652372619514;
   expected_results[52] =  0.34508248652372619514;
   expected_results[53] = -0.18748886006595252840;
   expected_results[54] =  0.04202636527909262010;
   expected_results[55] =  0.01305019940644617724;
   expected_results[56] = -0.27751259202078676358;
   expected_results[57] = -0.08413108928807125353;
   expected_results[58] =  0.03928220651181194012;
   expected_results[59] =  0.05901165657804177678;
   expected_results[60] =  0.00348155518956557580;
   expected_results[61] = -0.04191579352767203781;
   expected_results[62] =  0.08689468846600580110;
   expected_results[63] =  0.01301250123477310644;
   expected_results[64] = -0.04032311897593984090;
   expected_results[65] =  0.07474403970895460159;
   expected_results[66] =  0.04786291183903680635;
   expected_results[67] = -0.02428828698861155641;
   expected_results[68] = -0.06001906385307194391;
   expected_results[69] = -0.00795068167375006553;
   expected_results[70] =  0.02643447567385441566;
   expected_results[71] = -0.05538408271236267694;
   expected_results[72] = -0.01250032551907871887;
   expected_results[73] =  0.03869181896686624267;
   expected_results[74] = -0.02261690245306222438;
   expected_results[75] = -0.02165135036998024759;
   expected_results[76] =  0.01854924814461966776;
   expected_results[77] =  0.04096041084703053864;
   expected_results[78] =  0.00775974745641534649;
   expected_results[79] = -0.01903462878624252730;
   expected_results[80] =  0.02327368288411046579;
   expected_results[81] =  0.00305755210718031091;
   expected_results[82] = -0.02907374717387719784;
   expected_results[83] =  0.00290880442387733175;
   expected_results[84] =  0.00849548530949226915;
   expected_results[85] = -0.00878876776373385508;
   expected_results[86] = -0.01695735115373797272;
   expected_results[87] = -0.00176665961306401976;
   expected_results[88] =  0.01255794680690700688;
   expected_results[89] = -0.00441659222550526767;
   expected_results[90] =  0.00096535883615447675;
   expected_results[91] =  0.01389581931260174506;
   expected_results[92] = -0.00040579843413544522;
   expected_results[93] = -0.00484279055650574299;
   expected_results[94] =  0.00119935104891564755;
   expected_results[95] =  0.00305927679687127392;
   expected_results[96] = -0.00063717027408990740;
   expected_results[97] = -0.00456747888272199080;
   expected_results[98] = -0.00002202658825579365;
   expected_results[99] = -0.00016728479105140539;
   expected_results[100] = -0.00312243468817884279;
   expected_results[101] =  0.00091345773396801701;
   expected_results[102] =  0.00222989309130521094;
   expected_results[103] =  0.00027151056242530401;

   bands = (double *)malloc(2 * NUM_BANDS * sizeof(double));
   weights = (double *)malloc(NUM_BANDS * sizeof(double));
   desired = (double *)malloc(NUM_BANDS * sizeof(double));
   h = (double *)malloc(300 * sizeof(double));

   if (bands == NULL || weights == NULL || desired == NULL || h == NULL)
   {
      printf("REMEZ TEST: Error allocating memory for bands.\n");
      free(bands);
      free(weights);
      free(desired);
      free(h);
      return EXIT_FAILURE;
   }

   desired[0] = 0;
   desired[1] = 1; 
   desired[2] = 0;
   desired[3] = 1;
   desired[4] = 0;

   weights[0] = 10;
   weights[1] = 1;
   weights[2] = 3;
   weights[3] = 1;
   weights[4] = 20;

   bands[0] = 0;
   bands[1] = 0.05;
   bands[2] = 0.1;
   bands[3] = 0.15;
   bands[4] = 0.18;
   bands[5] = 0.25;
   bands[6] = 0.3;
   bands[7] = 0.36;
   bands[8] = 0.41;
   bands[9] = 0.5;

   status = remez(h, NUM_TAPS, NUM_BANDS, bands, desired, weights, BANDPASS);
   if (status != REMEZ_SUCCESS)
   {
      printf("REMEZ TEST: remez returned status %d\n", status);
      free(bands);
      free(weights);
      free(desired);
      free(h);
      return EXIT_FAILURE;
   }

   for (i=0; i<NUM_TAPS; i++)
   {
       if (!doubles_are_equal(h[i], expected_results[i]))
       {
           printf("REMEZ TEST: %.8e does not match expected result %.8e at "
               "index %d\n", h[i], expected_results[i], i);
           results_matched = 0;
       }
   }

   if (!results_matched)
   {
      printf("REMEZ TEST: Error: actual remez results didn't match expected "
          "results.\n");
      free(bands);
      free(weights);
      free(desired);
      free(h);
      return EXIT_FAILURE;
   }

   printf("REMEZ TEST: Remez results matched expected results. Test passed.\n");

   free(bands);
   free(weights);
   free(desired);
   free(h);

   return EXIT_SUCCESS;
}

